module dmdscript.pretty;

import dmdscript.script;
import dmdscript.program;
import dmdscript.errmsgs;

import dmdscript.dobject;
import dmdscript.value;
import dmdscript.darray;

import dmdscript.dnative;
import dmdscript.property;

import dmdscript.dglobal; // for readln and friends

import std.traits;
import std.typecons;

public import std.variant;

class ScriptEngine {
	this() {
		program = new Program;
	}

	// would be nice to have a global property that fetches an opDispatch wrapper for the script object...

	void compile(string data) {
		program.compile(null, data, null);
	}

	void execute() {
		program.execute(null);
	}

	final Variant call(T...)(string scriptFunctionName, T t) {
		Variant v;

		Value[] args;
		foreach(i; t) {
			Value arg;
			toScriptValue(i, &arg);
			args ~= arg;
		}

		Value value;
		value.putVundefined();

		Dobject sfun = program.callcontext.global.Get(scriptFunctionName).toObject();

		if(sfun.Call(program.callcontext,
			     program.callcontext.global,
			     &value,
			     args))
			throw new Exception("Script threw an error");
		else {
			//writefln("returned: %s", value.toString());

			return v;
		}
	}


	void addFunction(alias T, string name = "")(ScriptObject parent = null) {
		Dobject obj;
		if(parent is null)
			obj = program.callcontext.global;
		else
			obj = parent.value.toObject();

		string scriptName;
		static if(name == "")
			scriptName = __traits(identifier, T);
			//static assert(0, "It can't deduce the name :(");
		else
			scriptName = name;
		.addFunction!(T)(scriptName, obj);
	}

	// Adds the default functions from DMDScript: print, println, readln, and getenv
	void addDefaultFunctions() {
	    NativeFunctionData[] nfd = [
	    {	"print", &Dglobal_print, 1 },
	    {	"println", &Dglobal_println, 1 },
	    {	"readln", &Dglobal_readln, 0 },
	    {	"getenv", &Dglobal_getenv, 1 }
	    ];

	    DnativeFunction.init(program.callcontext.global, nfd, DontEnum);
	}

	ScriptObject globalObject() {
		return new ScriptObject(program.callcontext.global);
	}

	private Program program;
}

private void addFunction(alias T)(string scriptName, Dobject obj) {
	void* scriptWrapper(Dobject pthis, CallContext *cc, Dobject othis, Value* ret, Value[] arglist) {
		if(arglist.length != ParameterTypeTuple!(T).length) {
			ret.putVundefined();
			ErrInfo* errinfo = new ErrInfo;
			return Dobject.RuntimeError(errinfo, "Wrong number of arguments to native function. Expected: %d, got %d", ParameterTypeTuple!(T).length, arglist.length);
		}

		try
		static if( is(ReturnType!(T) == void ) ) {
			ret.putVundefined();
			callFunctionFromArgList!(T, ParameterTypeTuple!(T))(arglist);
		} else {
			auto retd = callFunctionFromArgList!(T, ParameterTypeTuple!(T))(arglist);
			toScriptValue(retd, ret);
		}
		catch (Exception e) {
			ret.putVundefined();
			ErrInfo* errinfo = new ErrInfo;
			errinfo.message = e.toString();
			return Dobject.RuntimeError(errinfo, "Native function threw: %s", e.toString());
		}
		return null;
	}

	// first arg is what object it belongs to btw
	DnativeFunction.initWithDelegate(obj, scriptName, &scriptWrapper, ParameterTypeTuple!(T).length, DontEnum);
}



class ScriptObject {
//	private Dobject obj;
	private Value* value;

	this(Dobject dobj = null) {
		value = new Value;
		if(dobj !is null)
			value.putVobject(dobj);
		else
			value.putVobject(new Dobject(Dobject.getPrototype()));
	}

	this(Value* v) {
		if( v is null)
			throw new Exception("null value");
		value = v;
	}

	ScriptObject opIndex(string i) {
		Value* v;
		v = value.Get(i);
		if(v is null)
			throw new Exception("no such index");
		return new ScriptObject(v);
	}

	ScriptObject opIndexAssign(T)(T w, string i) if( !isFunction!(T) /* HACK */ ) {
		Value* v = new Value;
		toScriptValue(w, v);
		value.Put(i, v);

		return this;
	}

	ScriptObject opIndexAssign(T)(T w, string i) if(isFunction!(T)) { // functions get special treatment
		Dobject obj;
		obj = value.toObject();

		static assert(0, "This currently won't work since it will make dmd fail an internal assert");
		addFunction!(w)(i, obj);

		return this;
	}
	/*
	void opDispatch(string name, T)(T t) if (!is(T: ScriptObject)) {
		toScriptValue(t, value);
	}
	*/

	/*
	ScriptObject opAssign(T)(T t) if (!is(T:ScriptObject)) {
		toScriptValue(t, value);
		return this;
	}
	*/
	ScriptObject opDispatch(string name)(ScriptObject o = null) {
		static if (name == "")
			static assert(0);
		if(o is null) {
			return opIndex(name);
		} else {
			return opIndexAssign(o, name);
		}
	}
/*
	Value* opDispatch(string name)(ScriptObject val = null) {
		if(val !is null) {
			Value * v = new Value;
			v.putVobject(val.obj);
			obj.Put(name, v, 0);
			return v;
		} else {
			auto v = obj.Get(name);
			if(v is null)
				throw new Exception("No such property: " ~ name);
			return v;
		}
	}
*/
}

template isFunction(T) {
	enum bool isFunction = inString(T.stringof, "function");
}

bool inString(string s, string find) {
	int pos = 0;
	foreach(c; s) {
		if(c == find[pos]) {
			pos++;
			if(pos == find.length)
				return true;
		} else
			pos = 0;
	}

	return false;
	/* // this does:
	if(s.indexOf("function") == -1)
		return false;
	else
		return true;
	*/
}

/**
	Puts a normal D value into a script value;

	If you give the ret argument, it saves it
	into there. If not, it allocates a new Value
	and returns it;
*/
Value* toScriptValue(T)(T retd, Value* ret = null) {
	if(ret is null)
		ret = new Value;

	static if(is(T == Value*)) // if we are given a script value, just use it directly
		Value.copy(ret, retd);
	else static if(is(T : ScriptObject)) // our ScriptObject is just a wrapper for Value, so unwrap it and call it done
		Value.copy(ret, retd.value);

	else static if(isNumeric!(T))
		ret.putVnumber(retd);
	else static if(isSomeString!(T))
		ret.putVstring(retd);
	else static if(isSomeChar!(T))
		ret.putVstring([retd]);
	else static if(isArray!(T)) {
		static if(is(Unqual!(T) == void[])) { // assuming void[] are string, since that seems most useful for scripts
			toScriptValue((cast(char[]) retd).idup, ret); // iduping it just for safety
		} else {
			auto o = new Darray();
			foreach(i, t; retd) { // gotta build the array object
				auto a = new Value;
				toScriptValue(t, a);
				o.Put(i, a, 0);
			}
			ret.putVobject(o);
		}
	}
	else static assert(0, "Can't convert to script value automatically");

	return ret;
}

T fromScriptValue(T)(Value arg) {
	static if(isIntegral!(T))
		return arg.toInt32();
	else static if(isFloatingPoint!(T))
		return arg.toNumber();
	else static if(is(T == bool))
		return arg.toBoolean();
	else static if(isSomeString!(T))
		return arg.toString();
	else static assert(0, "I don't know how to get that D type out of the script value");
}

string intToString(int i) {
	if(i == 0)
		return "0";
	string r;
	do {
		int v = i % 10;
		i /= 10;
		r ~= cast(char) (v + '0');
	} while (i != 0);

	string ret;
	for(int a = r.length - 1; a >= 0; a--)
		ret ~= r[a];

	return ret;
}

auto callFunctionFromArgList(alias F, T...)(Value[] arglist) {
	string helper(int length) {
		string ret;
		for(int a = 0; a < length; a++) {
			if(a > 0)
				ret ~= ", ";
			ret ~= "fromScriptValue!(T["~intToString(a)~"])(arglist["~intToString(a)~"])";
		}

		return ret;
	}

	static if(T.length == 0)
		return F();
	else
		return mixin("F(" ~ helper(T.length) ~ ")");
}

